;; Generated package description from cider.el  -*- mode: lisp-data; no-byte-compile: t; lexical-binding:t -*-
(define-package "cider" "1.22.0snapshot0.20260416.192551" "Clojure Interactive Development Environment that Rocks" '((emacs "28") (clojure-mode "5.19") (compat "30") (parseedn "1.2.1") (queue "0.2") (spinner "1.7") (seq "2.22") (sesman "0.3.2") (transient "0.4.1")) :commit "8382ff70b7646bc354603fe7d11407435b3ac732" :authors '(("Tim King" . "kingtim@gmail.com") ("Phil Hagelberg" . "technomancy@gmail.com") ("Bozhidar Batsov" . "bozhidar@batsov.dev") ("Artur Malabarba" . "bruce.connor.am@gmail.com") ("Hugo Duncan" . "hugo@hugoduncan.org") ("Steve Purcell" . "steve@sanityinc.com")) :maintainer '("Bozhidar Batsov" . "bozhidar@batsov.dev") :keywords '("languages" "clojure" "cider") :url "https://www.github.com/clojure-emacs/cider")
