((nil . ((eval . (customize-set-variable 'cider-path-translations
                                         (list
                                          (cons "/src" (clojure-project-dir))
                                          (cons "/root/.m2" (concat (getenv "HOME") "/.m2"))))))))
